from plantstar_shared import syscon_json

# Converts the output dictionary into a string through syscon_json.dumps, then encodes it into bytes so that it can be sent through the socket, and parsed later on
def convert_dictionary_to_bytes(output_dictionary, encoding_format="ASCII"):
    output_dictionary_as_string = syscon_json.dumps(output_dictionary)
    output_dictionary_as_bytes = output_dictionary_as_string.encode(encoding_format)

    return output_dictionary_as_bytes



import json
import datetime
import uuid
import pytz
from dateutil.tz import tzutc
from psycopg2._range import DateTimeTZRange

from plantstar_shared.syscon_image_field import SysconImageFieldFile


class SysconEncoder(json.JSONEncoder):
    def default(self, obj):
        if obj is None:
            return None

        if hasattr(obj, "tzinfo"):
            # This is here for a weird bug that was happening
            if obj.tzinfo == tzutc:
                tzinfo = "UTC"
            elif isinstance(obj.tzinfo, datetime.timezone):
                if obj.tzinfo.tzname(None) in ["+0000", "UTC"]:
                    tzinfo = "UTC"
                else:
                    raise Exception("Fixed offset is not utc, we need to figure out how to solve this issue.")
            else:
                tzinfo = str(obj.tzinfo)
        else:
            tzinfo = "UTC"

        if isinstance(obj, datetime.datetime):
            return {
                "__type__": "datetime",
                "year": obj.year,
                "month": obj.month,
                "day": obj.day,
                "hour": obj.hour,
                "minute": obj.minute,
                "second": obj.second,
                "microsecond": obj.microsecond,
                "tzinfo": tzinfo
            }
        elif isinstance(obj, datetime.date):
            return {
                "__type__": "date",
                "year": obj.year,
                "month": obj.month,
                "day": obj.day
            }
        elif isinstance(obj, datetime.time):
            return {
                "__type__": "time",
                "hour": obj.hour,
                "minute": obj.minute,
                "second": obj.second,
                "microsecond": obj.microsecond,
                "tzinfo": tzinfo
            }
        elif isinstance(obj, uuid.UUID):
            return str(obj)
        elif isinstance(obj, DateTimeTZRange):
            return {
                "__type__": "datetime_range",
                "lower": obj.lower,
                "upper": obj.upper,
                "bounds": obj._bounds
            }
        elif isinstance(obj, SysconImageFieldFile):
            if obj:
                return {
                    "__type__": "SysconImageFieldFile",
                    "path": obj.src_path
                }
            else:
                return ""
        else:
            return super(SysconEncoder, self).default(obj)


class SysconDecoder(json.JSONDecoder):
    def __init__(self):
        json.JSONDecoder.__init__(self, object_hook=self.decode_special_types)

    def decode_special_types(self, dictionary):
        if "__type__" not in dictionary:
            return dictionary

        type_name = dictionary.pop("__type__")

        if "tzinfo" in dictionary:
            # This is here for a weird bug that was happening
            if dictionary["tzinfo"] is None or dictionary["tzinfo"] in ["None", "tzlocal()", "tzutc()"]:
                dictionary["tzinfo"] = "UTC"

            dictionary["tzinfo"] = pytz.timezone(dictionary["tzinfo"])

        if type_name == "datetime":
            return datetime.datetime(**dictionary)
        elif type_name == "date":
            return datetime.date(**dictionary)
        elif type_name == "time":
            return datetime.time(**dictionary)
        elif type_name == "uuid":
            return uuid.UUID(**dictionary)
        elif type_name == "datetime_range":
            return DateTimeTZRange(**dictionary)
        else:
            dictionary["__type__"] = type_name
            return dictionary


class SysconSerializer:
    def dumps(self, obj):
        return json.dumps(obj, cls=SysconEncoder, separators=(",", ":")).encode("utf-8")

    def loads(self, string):
        return json.loads(string.decode("utf-8"), cls=SysconDecoder)


def dumps(obj, **kwargs):
    return json.dumps(obj, cls=SysconEncoder, **kwargs)


def dump(obj, file_pointer, **kwargs):
    return json.dump(obj, file_pointer, cls=SysconEncoder, **kwargs)


def loads(string, **kwargs):
    return json.loads(string, cls=SysconDecoder, **kwargs)


def load(file_pointer, **kwargs):
    return json.load(file_pointer, cls=SysconDecoder, **kwargs)



from plantstar_shared import syscon_json

# Converts bytes into a decoded string, then into a dictionary using syscon_json.loads
def convert_bytes_to_dictionary(output_dictionary_as_bytes, decoding_format="ASCII"):
    output_dictionary_as_string = output_dictionary_as_bytes.decode(decoding_format)
    output_dictionary = syscon_json.loads(output_dictionary_as_string)

    return output_dictionary



class RawDataProcessorInterfaceActions:
    GET_DRIVER_VERSION_STRING = "get_driver_version_string"
    GET_INPUT_SIGNAL_DICTIONARIES = "get_input_signal_dictionaries"
    GET_MOCK_RAW_DATA_PROCESSOR_STATUS = "get_mock_raw_data_processor_status"
    GET_VERSION_STRING = "get_version_string"
    UPDATE_ANALOG_INPUTS = "update_analog_inputs"
    UPDATE_DIGITAL_INPUT_DEBOUNCE = "update_digital_input_debounce"
    UPLOAD_NEW_DRIVER_SETTINGS = "upload_new_driver_settings"



import datetime
import time
import array

from django.utils.timezone import now

TEN_MILLISECONDS = .01


class MockRawDataProcessor:
    class EmptyInputSignalDictionaries(Exception):
        pass

    def __init__(
        self, cycle_time_in_milliseconds=30000, number_of_digital_inputs=16, number_of_digital_outputs=8, number_of_analog_inputs=32, debounce_raw_buffer_size=255,
        analog_value_min=50, analog_value_max=2, analog_value_step=0.5
    ):
        self.next_value_swap_in_seconds_timedelta = datetime.timedelta(milliseconds=cycle_time_in_milliseconds / 2)
        self.next_value_swap_datetime = now() + self.next_value_swap_in_seconds_timedelta

        self.number_of_digital_inputs = number_of_digital_inputs
        self.number_of_digital_outputs = number_of_digital_outputs
        self.number_of_analog_inputs = number_of_analog_inputs
        self.debounce_raw_buffer_size = debounce_raw_buffer_size
        self.analog_value_min = analog_value_min
        self.analog_value_max = analog_value_max
        self.analog_value_step = analog_value_step

        self.last_digital_input_debounce_array = array.array("B", [0 for _ in range(self.number_of_digital_inputs)])
        self.last_digital_output_array = array.array("B", [0 for _ in range(self.number_of_digital_outputs)])
        self.last_analog_filter_array = array.array("B", [0 for _ in range(self.number_of_analog_inputs)])
        self.last_analog_gain_array = array.array("B", [0 for _ in range(self.number_of_analog_inputs)])

        self.state_on_or_off_value = False
        self.current_analog_value = analog_value_max
        self.current_analog_step_direction = -1
        self.cold_end_value = 20

    def check_data_length_and_limit(self, input_data_buffer, data_buffer_length_limit, data_buffer_int_limit):
        if len(input_data_buffer) == data_buffer_length_limit:
            for i in range(data_buffer_length_limit):
                if input_data_buffer[i] < 0 or input_data_buffer[i] > data_buffer_int_limit:
                    raise ValueError(f"Incorrect value passed into array.array object. It must be 0 through {data_buffer_int_limit}.")
        else:
            raise ValueError(
                "Expected a memory view with {0} items, {1}{2} were received.".format(
                    data_buffer_length_limit, "only " if len(input_data_buffer) < data_buffer_length_limit else "", len(input_data_buffer)
                )
            )

    def upload_new_driver_settings(self, data_type_string, data_buffer):
        if data_type_string == "digital_outputs":
            data_buffer_len = self.number_of_digital_outputs
            data_buffer_int_limit = 1
            self.check_data_length_and_limit(data_buffer, data_buffer_len, data_buffer_int_limit)

            if self.last_digital_output_array != data_buffer:
                self.last_digital_output_array = data_buffer
        elif data_type_string == "analog_gains":
            data_buffer_len = self.number_of_analog_inputs
            data_buffer_int_limit = 3
            self.check_data_length_and_limit(data_buffer, data_buffer_len, data_buffer_int_limit)

            if self.last_analog_gain_array != data_buffer:
                self.last_analog_gain_array = data_buffer
        elif data_type_string == "analog_filters":
            data_buffer_len = self.number_of_analog_inputs
            data_buffer_int_limit = 15
            self.check_data_length_and_limit(data_buffer, data_buffer_len, data_buffer_int_limit)

            if self.last_analog_filter_array != data_buffer:
                self.last_analog_filter_array = data_buffer
        elif data_type_string == "digital_input_debounce":
            data_buffer_len = self.number_of_digital_inputs
            data_buffer_int_limit = self.debounce_raw_buffer_size - 1
            self.check_data_length_and_limit(data_buffer, data_buffer_len, data_buffer_int_limit)

            if self.last_digital_input_debounce_array != data_buffer:
                self.last_digital_input_debounce_array = data_buffer
        else:
            if type(data_type_string) == str:
                raise ValueError(
                    f"An Incorrect value of {data_type_string} was passed. This function expects \"digital_outputs\", \"analog_gains\", or \"analog_filters\"."
                )
            else:
                raise TypeError(
                    f"Failed to pass a recognizable type of data. {data_type_string} was passed in with a {type(data_type_string)} data type. This function expects a string "
                    f"of \"digital_outputs\", \"analog_gains\", or \"analog_filters\"."
                )

    def get_input_signal_dictionaries(self):
        time.sleep(TEN_MILLISECONDS)

        the_now = now()

        if self.next_value_swap_datetime <= the_now:
            self.state_on_or_off_value = not self.state_on_or_off_value
            self.next_value_swap_datetime = the_now + self.next_value_swap_in_seconds_timedelta

            if self.state_on_or_off_value:
                self.current_analog_value = self.analog_value_max
                self.current_analog_step_direction = -1

        return_dictionary = {
            1: {
                "digitals": {str(state_index + 1): int(self.state_on_or_off_value) for state_index in range(16)},
                "analogs": {str(state_index + 1): self.current_analog_value for state_index in range(24)},
                "cold_end": {str(state_index + 1): self.cold_end_value for state_index in range(24)},
                "event_unix_timestamp": the_now.timestamp()
            }
        }

        self.step_current_analog_value()

        return return_dictionary

    def get_driver_version_string(self):
        return "Mock"

    def get_version_string(self):
        return "Mock"

    def step_current_analog_value(self):
        if self.current_analog_step_direction == -1:
            step_min_max_function = max
            ceiling_or_floor_value = self.analog_value_min
        else:
            step_min_max_function = min
            ceiling_or_floor_value = self.analog_value_max

        new_analog_value = step_min_max_function(self.current_analog_value + self.analog_value_step * self.current_analog_step_direction, ceiling_or_floor_value)

        if new_analog_value == ceiling_or_floor_value:
            self.current_analog_step_direction *= -1

        self.current_analog_value = new_analog_value
